/*
  Copyright (C) 1998 Ralf Wiebicke

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You'll find the GPL at http://www.gnu.org/copyleft/gpl.html
*/

package de.rw7.htmltools;

import java.net.URL;
import java.net.MalformedURLException;
import java.net.UnknownHostException;
import java.io.FileNotFoundException;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

/**
  Wraps an url and holds additional information about the url.
  Each url is associated to exactly one LinkFile.
*/
public class LinkFile extends LinkTag
{
  /**
    The associated url.
  */  
  URL url;
  
  /**
    Holds the status of the url.
    <UL>
    <LI>0: vakant, yet to be checked</LI>
    <LI>1: lookup, currently checking</LI>
    <LI>2: found, url is valid</LI>
    <LI>3: absent, url is not valid</LI>
    </UL>
  */
  int state;
  
  public final int getState()
  {
    return state;
  }

  /**
    A description of the error, causing the url to be absent.
  */
  protected String error;

  /**
    Whether the url is external to the root url. External means,
    that it's not in the same directory or one of the subdirectorys
    of the root url. Here, the root url given to the constructor is used.
  */
  boolean external;

  /**
    Contains all anchors of this url. The Hashtable maps from the 
    anchors name to the anchors LinkAnchor-Object.
    @see LinkAnchor
  */
  protected Hashtable anchors;
  
  public final Hashtable getAnchors()
  {
    return (anchors!=null)?(Hashtable)anchors.clone():null;
  }
  
  protected String rootHost, rootDir;

  /**
    The formatted url, relative to the root url.
  */  
  protected String urlS;
  
  public LinkFile(URL url, String rootHost, String rootDir) 
  {
    this.state=0;
    this.url=url;
    this.external=
      !rootHost.equals(url.getHost()) ||
      !url.getFile().startsWith(rootDir) ;
    this.rootHost=rootHost;
    this.rootDir=rootDir;
    this.referrers=new Vector();
  }
  
  public LinkFile(String urlS, String error) 
  {
    this.state=3;
    this.error=error;
    this.external=false;
    this.referrers=new Vector();
    this.urlS=urlS;
  }
  
  public final String toString()
  {
    if(urlS!=null) 
      return urlS;
    else
      if(rootHost.equals(url.getHost()))
      {
        String file=url.getFile();
        if(file.startsWith(rootDir))
          return urlS=file.substring(rootDir.length());
        else
          return urlS=url.getFile();
      }
      else
        return urlS=url.toString();
  }
  
  public final String getError()
  {
    return error;
  }
    
  synchronized final void addReferrer(LinkFile referrer) 
  {
    if(state!=2) 
      referrers.addElement(referrer);
  }
  
  synchronized final void toLookup()
  {
    if(state!=0) throw new IllegalArgumentException();
    state=1;
  }
    
  synchronized final void toFound(LinkFrontend fe)
  {
    if(state!=1) throw new IllegalArgumentException();
    state=2;
    referrers=null;

    if(anchors!=null)
      for(Enumeration i=anchors.elements(); i.hasMoreElements(); )
      {
        LinkAnchor la=(LinkAnchor)i.nextElement();
        if(!la.isFound()) fe.onAbsent(la);
      }
  }
    
  synchronized final void toAbsent(Exception e)
  {
    //System.err.println("absent "+toString()+" - "+e.toString());
    if(state!=1) throw new IllegalArgumentException();

    if(e instanceof HttpException)
      error="http response: "+((HttpException)e).code+" "+e.getMessage();
    else if(e instanceof UnknownHostException)
      error="unknown host: \""+e.getMessage()+"\"";
    else if(e instanceof FileNotFoundException)
      error="file not found: \""+e.getMessage()+"\"";
    else if(e instanceof MalformedURLException)
      error="malformed url: \""+e.getMessage()+"\"";
    else
      error=e.toString();

    state=3;
  }
  
  synchronized final LinkAnchor anchorRequired(String anchor, LinkFrontend fe)
  {
    if(!external)
    {
      if(anchors==null) anchors=new Hashtable();
      LinkAnchor la;
      if((la=(LinkAnchor)(anchors.get(anchor)))==null) 
      { 
        anchors.put(anchor, la=new LinkAnchor(this, anchor)); 
        if(state==2) fe.onAbsent(la);
      }
      return la;
    }
    else
      return null;
  }
  
  synchronized final void anchorFound(String anchor)
  {
    if(external) throw new IllegalArgumentException();
    if(anchors==null) anchors=new Hashtable();

    LinkAnchor la=(LinkAnchor)(anchors.get(anchor));
    if(la==null)
      anchors.put(anchor, new LinkAnchor(this, anchor, true)); 
    else
      la.toFound(); 
  }  
  
}
