/*
  Copyright (C) 1998 Ralf Wiebicke

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You'll find the GPL at http://www.gnu.org/copyleft/gpl.html
*/

package de.rw7.htmltools;

import java.io.IOException;
import java.net.URL;
import java.net.URLConnection;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import de.rw7.token.HtmlConsumer;
import de.rw7.token.FastInput;
import de.rw7.token.HtmlTokenizer;

/**
  This class actually does the work. It runs over an entire website,
  or a subdirectory, fishing out all broken links in HTML and CSS 
  code. It also holds the information, which urls have been checked 
  already, and which not.<P>
  Checking is done multithreaded. For each thread an instance of
  the inner class T is created.
*/
public class LinkSpider
{
  /**
    The associated frontend.
  */
  protected LinkFrontend fe;
  
  /**
    The root url to be checked.
  */
  protected URL root;
  
  /**
    Tells the spider, to check links, which are not in or
    under the directory of the root url.
  */
  protected boolean verifyExternals;

  /**
    Holds all urls checked up to now.
  */
  protected Hashtable pool;
  
  /**
    For debugging only.
  */
  public final Hashtable getPool() 
  { 
    return (Hashtable)pool.clone(); 
  };
  
  /**
    Holds all urls not yet checked.
  */
  protected Vector vakant;
  
  /**
    For debugging only.
  */
  public final Enumeration getVakant() 
  { 
    return vakant.elements(); 
  };

  /**
    Holds all defined target frame names. Maps the target name
    to the corresponding LinkTarget-Object.
    @see de.rw7.htmltools.LinkTarget
  */
  protected Hashtable targets;
  
  /**
    For debugging only.
  */
  public final Hashtable getTargets() 
  { 
    return (Hashtable)targets.clone(); 
  };
  
  /*
    Holds the host of the root url.
  */
  protected String rootHost;
  
  /**
    Holds the directory of the root url.
  */
  protected String rootDir;
  
  protected HtmlTokenizer ht;
  
  /**
    Holds all threads of the spider.
  */
  protected T[] threads;
  
  /** 
    The number of threads not waiting due to lack of work. 
    If this becomes zero, the spider exits normally.
  */
  protected int runningThreads;
    
  /**
    The constructor of the spider. Does not start the spider.
    @param lv the accociated frontend window.
    @param src the root url.
    @param verifyExternals whether to check url not in or under
                           the root urls directory
    @param nthreads the number of threads to be created.
  */
  
  public LinkSpider(LinkFrontend fe,
                    URL root,
                    boolean verifyExternals,
                    int nthreads)
  {
    this.fe=fe;
    this.root=root;
    this.verifyExternals=verifyExternals;
    this.ht=new HtmlTokenizer(0); setTags(ht); //ht.printTree();
    this.threads=new T[nthreads];
    if(root==null) throw new IllegalArgumentException();
  }
  
  /**
    Starts the spider. From the moment of calling this
    method, the associated Frontend has to be able to
    handle calls to onAbsent etc.
  */
  public void startVerify()
  {
    pool=new Hashtable();
    vakant=new Vector();
    runningThreads=threads.length;

    rootHost=root.getHost();
    rootDir=extractDir(root.getFile());
    System.err.println("root("+rootHost+","+rootDir+")");

    urlRequiredGet(root,null);

    targets=new Hashtable();
    
    /*
    try
    {
      Reader r=new InputStreamReader(new URL(root).openStream());
      int c; while((c=r.read())>=0) System.err.print((char)c);
    }
    catch(IOException e) { System.err.println(e); }
    */
    
    for(int i=0; i<threads.length; i++) threads[i]=new T(i);
  }

  /**
    Stops the spider abnormally. This is called by the frontend on
    user interaction.
  */  
  public final void stop()
  {
    synchronized(pool)
    {
      //System.err.println(Thread.currentThread().getName()+" STOP");
      if(threads!=null)
      {
        for(int i=0; i<threads.length; i++) 
          if(threads[i]==Thread.currentThread())
            throw new IllegalArgumentException();
        for(int i=0; i<threads.length; i++) 
        {
          threads[i].stop();
          threads[i]=null;
        }
        fe=null; // WORKAROUND for green threads
      }    
    }
  }

  protected final URL urlRequiredGet(URL u, LinkFile referrer)
  {
    String ref=u.getRef();

    try
    {
      if(bugEqualsPresent && u.getRef()!=null)
        u=new URL(u.getProtocol(),u.getHost(),u.getPort(),u.getFile());
      if(bugHostPresent && "file".equals(u.getProtocol()) && !"".equals(u.getHost()))
        u=new URL(u.getProtocol(),"",u.getPort(),u.getFile());
    }
    catch(MalformedURLException e) {System.out.println(e);return null;} // should not happen
    
    LinkFile lf;
    synchronized(pool)
    {
      lf=(LinkFile)pool.get(u);

      if(lf==null) 
      { 
        lf=new LinkFile(u, rootHost, rootDir);
        if(lf.external&&!verifyExternals) return u;
        pool.put(u, lf); 
        vakant.addElement(lf);
        pool.notify(); 
      }
    }

    if(referrer!=null) 
      lf.addReferrer(referrer);

    if(ref!=null) 
    {
      LinkAnchor la=lf.anchorRequired(ref, fe);
      if(la!=null&&referrer!=null) 
        la.addReferrer(referrer);
    }
    return u;
  }

  protected final LinkFile urlLookup(int count) throws InterruptedException
  {
    LinkFile ut;
    synchronized(pool)
    {
      
      while(vakant.size()==0)
      {
        //System.err.println(Thread.currentThread().getName()+" wait");
        runningThreads--;
        if(runningThreads<=0)
        {
          //System.err.println(Thread.currentThread().getName()+" KICK");
          ((T)(Thread.currentThread())).doOnEnd=true;
          pool.notifyAll();
          return null;
        }  
        fe.onWaiting(count);
        pool.wait(); 
        if(runningThreads<=0)
        {
          //System.err.println(Thread.currentThread().getName()+" exit");
          return null;
        }
        runningThreads++;
        //System.err.println(Thread.currentThread().getName()+" unwait");
      }

      ut=(LinkFile)vakant.lastElement(); 
      //System.err.println(count+"found ut");
      vakant.setSize(vakant.size()-1);
      ut.toLookup();
    }
    fe.onLookup(ut,count);
    return ut;
  }
  
  protected final void urlAbsent(LinkFile ut, Exception e)
  {
    ut.toAbsent(e);
    fe.onAbsent(ut);
  }

  protected int targetNumber=0;
  
  protected final void target(LinkFile lf, String target, boolean definition)
  {
    //System.err.println("TARGET "+(definition?"DEF ":"ref ")+target);

    LinkTarget ta;
    boolean update;
    synchronized(targets)
    {
      if((ta=(LinkTarget)targets.get(target))==null)
      {
        targets.put(target, ta=new LinkTarget(lf, target, targetNumber++, definition));
        update=true;
      }  
      else
        update=ta.updateState(lf, definition);
    }
    if(update) fe.onTarget(ta);
  }
  
  private static final String extractDir(String filename)
  {
    int i=filename.lastIndexOf((int)'/');
    return (i<0) ? filename : filename.substring(0,i+1);
  }
  
  private static boolean bugEqualsPresent;
  private static boolean bugHostPresent;
  
  static
  {
    try
    {
      if( bugEqualsPresent=(!(new URL("file:///meag#1")).equals(new URL("file:///meag#2"))) )
        System.err.println("bug in java.net.URL.equals() detected, workaround enabled, bugid:4052976 (fixed in 1.2beta4, not yet in 1.1.x)");
      
      if( bugHostPresent=(!"".equals(new URL("file://localhost/d:/source/html/").getHost())))
        System.err.println("bug in java.net.URL.getHost() detected, workaround enabled, bugid:25701");
    }
    catch(IOException e) {System.err.println(e);}
  }
  
  private static final int TARGET = 4096;
  
  private static final void setTags(HtmlTokenizer ht)
  {
    String[] aTargetHrefName = { "TARGET","HREF","NAME" };
    ht.addOpeningTag("A",    0+TARGET, aTargetHrefName);
    
    String[] aTargetHref = { "TARGET", "HREF" };
    ht.addOpeningTag("BASE", 1+TARGET, aTargetHref);
    
    String[] aSrcName = { "SRC", "NAME" };
    ht.addOpeningTag("FRAME",2, aSrcName);
    
    ht.addTag("STYLE", 10, null);

    String[] aApplet = { "SRC", "ARCHIVE" };
    ht.addOpeningTag("APPLET",  11, aApplet);

    ht.addOpeningTag("AREA", 21+TARGET, aTargetHref);
    
    String[] aHref = { "HREF" };
    ht.addOpeningTag("LINK",    21, aHref);
    
    String[] aImg = { "SRC", "LOWSRC", "USEMAP", "DYNSRC", "VRML" };
    ht.addOpeningTag("IMG",     25, aImg);
    
    String[] aSrc = { "SRC" };
    ht.addOpeningTag("IFRAME",  21, aSrc);
    ht.addOpeningTag("EMBED",   21, aSrc);
    ht.addOpeningTag("SOUND",   21, aSrc);
    ht.addOpeningTag("BGSOUND", 21, aSrc);
    ht.addOpeningTag("INPUT",   21, aSrc);
    ht.addOpeningTag("SCRIPT",  21, aSrc);

    String[] aBg = { "BACKGROUND" };
    ht.addOpeningTag("BODY",    21, aBg);
    ht.addOpeningTag("TABLE",   21, aBg);
    ht.addOpeningTag("TH",      21, aBg);
    ht.addOpeningTag("TD",      21, aBg); 

    String[] aSrcBg = { "SRC", "BACKGROUND" };
    ht.addOpeningTag("LAYER",   22, aSrcBg);
    ht.addOpeningTag("ILAYER",  22, aSrcBg);
    
    String[] aCite = { "CITE" };
    ht.addOpeningTag("INS",     21, aCite);
    ht.addOpeningTag("DEL",     21, aCite);
    ht.addOpeningTag("Q",       21, aCite);
    ht.addOpeningTag("ACRONYM", 21, aCite);
    
    String[] aTargetAction = { "TARGET", "ACTION" };
    ht.addOpeningTag("FORM",    21+TARGET, aTargetAction);
    
    String[] aAction = { "ACTION" };
    ht.addOpeningTag("ISINDEX", 21, aAction); 
  }
    
  class T extends Thread implements de.rw7.token.HtmlConsumer
  {
    private int count;
    boolean doOnEnd=false;

    T(int count) 
    {
      //System.err.println(count+"created");
      super("Link Verify Spider #"+count);
      this.count=count;
      setPriority(Thread.MIN_PRIORITY);
      start();
    }
  

    private URL refURL, refBASE;
    private LinkFile refTAG;

    public void run()
    {
      FastInput fi=new FastInput();
      try 
      {
    
      while((refTAG=urlLookup(count))!=null)
      {
        refBASE=refURL=refTAG.url;
      
        try
        {
          int contenttype=0;
          if(refURL.getFile().endsWith(".html")||
             refURL.getFile().endsWith(".shtml")||
             refURL.getFile().endsWith(".htm"))
          {
            contenttype=1;
          }
          else if(refURL.getFile().endsWith(".css"))
          {
            contenttype=2;
          }
        

          URLConnection uc=refURL.openConnection();
          HttpURLConnection uch=
            (uc instanceof HttpURLConnection)?(HttpURLConnection)uc:null;

          if(uch!=null)
            if(refTAG.external||contenttype==0) uch.setRequestMethod("HEAD");

          uc.connect();

          if(uch!=null) 
            if(uch.getResponseCode()!=200)
              throw new HttpException(uch.getResponseCode(), uch.getResponseMessage());

          String tp=uc.getContentType();

          // workaround for incompatibility in IE40
          if(tp==null) throw new IOException("URLConnection.getContentType()==null");
        
          if(!refTAG.external && contenttype==0)
          {
            if(tp.startsWith("text/html"))
              contenttype=1;
            else
            if(tp.startsWith("text/css"))
              contenttype=2;
            if(uch!=null && contenttype>0)
            {
              uch.disconnect();
              //System.err.println(count+"reconnecting "+refTAG.getURLS(ls));
              uc=refURL.openConnection();
              if(uc==uch) throw new IllegalArgumentException();
              uch=(HttpURLConnection)uc;
            }
          }
          
          if(!refTAG.external && contenttype>0)
          {
            //System.err.println(count+"parsing "+refTAG.getURLS(ls));
            fi.open(uc.getInputStream());
            switch(contenttype)
            {
            case 1:
              try { ht.read(fi,this); } catch(FastInput.EndException e){};
              break;
            case 2:
              //System.err.println(".css on: "+refTAG.getURLS(ls));
              try { readCss(fi); } catch(FastInput.EndException e){};
              //System.err.println(".css off: "+refTAG.getURLS(ls));
              break;
            default: throw new IllegalArgumentException();
            }
            fi.close();
          }
        }
        catch(IOException e) { urlAbsent(refTAG,e); continue; }
        catch(SecurityException e) { urlAbsent(refTAG,e); continue; }
      
        //System.err.println(count+"found "+refTAG.getURLS(ls));
        refTAG.toFound(fe);
      }
      } 
      catch (InterruptedException ex) {System.err.print(count);ex.printStackTrace();}

      for(int i=0; i<threads.length; i++) 
        threads[i]=null;
        
      if(doOnEnd) 
        fe.onEnd();
    }
  
    private final void urlRequired(String spec, LinkFile referrer)
    {
      if(spec==null) return;
      try
      {
        URL u=urlRequiredGet(new URL(refBASE,LinkTag.replaceBackslashes(spec)), referrer);
      }
      catch(MalformedURLException e)
      {
        LinkFile ut=new LinkFile(spec, e.toString());
        if(referrer!=null) ut.addReferrer(referrer);
        fe.onAbsent(ut);
      };
    }

    private final void urlRequiredBase(String spec, LinkFile referrer)
    {
      if(spec==null) return;
      try
      {
        URL u=urlRequiredGet(new URL(refBASE,LinkTag.replaceBackslashes(spec)), referrer);
        refBASE=u;
      }
      catch(MalformedURLException e)
      {
        LinkFile ut=new LinkFile(spec, e.toString());
        if(referrer!=null) ut.addReferrer(referrer);
        fe.onAbsent(ut);
      };
    }

    public final int onOpeningTag(int code, String tag, String[] attr)
    { 
      //System.err.print("O("+code+")"+tag+"-");
      //for(int i=0; i<attr.length; i++) System.err.print(attr[i]+"|");
      //System.err.println();
      
      int p=0;
      
      if(code>=TARGET) 
      { 
        code-=TARGET; 
        if(attr[p]!=null) target(refTAG, attr[p],false); 
        p++;
      }

      switch(code)
      {
      case 0: //A
        urlRequired(attr[p++], refTAG);
        if(attr[p]!=null) refTAG.anchorFound(attr[p]);
        return 0;
        
      case 25:
        urlRequired(attr[p++], refTAG);
        urlRequired(attr[p++], refTAG);
        urlRequired(attr[p++], refTAG);
      case 22:
        urlRequired(attr[p++], refTAG);
      case 21:
        urlRequired(attr[p++], refTAG);
        return 0;
        
      case 1: //BASE
        //System.err.println("BASE :"+attr[p]);
        urlRequiredBase(attr[p],refTAG);
        return 0;
        
      case 2: //FRAME
        urlRequired(attr[p++], refTAG);
        if(attr[p]!=null) target(refTAG, attr[p],true);
        return 0;
        
      case 10: //CSS
        return HtmlTokenizer.MODE_PARSE_TEXT;
        
      case 11: //APPLET
        urlRequired(attr[p++], refTAG);
        String s=attr[p++];
        if(s!=null)
        {
          //System.err.println("ARCHIVE: >"+s+"<");
          int front=0;
          while(front<s.length())
          {
            while(s.charAt(front)==' ') front++;
            int end=s.indexOf((int)',', front);
            if(end<0) end=s.length();
            int nextfront=end+1;
            while(s.charAt(end-1)==' ') end--;
            if(front+1<end) 
            {
              urlRequired(s.substring(front,end), refTAG);
              //System.err.println("archive: >"+s.substring(front,end)+"<");
            }
            front=nextfront;
          }
        }
        return 0;
        
      default: throw new IllegalArgumentException(String.valueOf(code));
      }
    }
  
    public final int onClosingTag(int code, String tag)
    { 
      return 0;
    }

    public final void onText(FastInput in) throws IOException, FastInput.EndException
    { 
      readCss(in);
    }

    private final void readCss(FastInput in) throws IOException, FastInput.EndException
    {
      StringBuffer b=new StringBuffer(10);
      //System.err.println("parsing css: "+refTAG.getURLS(ls));
    
      while(in.skipUntil((byte)'<',(byte)'u')=='u')
      {
        if(in.read()!='r') continue;
        if(in.read()!='l') continue;
        if(in.read()!='(') continue;
        char buf;
        while((buf=in.read())!=')') b.append(buf);
        //System.err.println("req css: "+b);  
        urlRequired(new String(b),refTAG);
        b.setLength(0);
      }
      in.pushBack();
    }
  
  } // class LinkSpider.T

}
