/*
  Copyright (C) 1998 Ralf Wiebicke

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You'll find the GPL at http://www.gnu.org/copyleft/gpl.html
*/

package de.rw7.token;

import java.io.InputStream;
import java.io.IOException;
import java.io.FileInputStream;

/**
  No concurrency for this and is;
*/

public class FastInput
{
  /**
    The input stream. Is null after close().
    @see #close()
  */  
  private InputStream is;

  private int nbuf;
  private byte[] buf;
  private byte[] buf_bak;

  /**
    Number of bytes read without the currently loaded buffer.
  */
  private int blocksread;
  
  /**
    Valid bytes in buffer. There may be read at least len-now 
    bytes without IO access.
  */
  private int len;
  
  /**
    The current byte in buffer, next to be read.
    The class ensures, that now is not zero, whenever pushBack is allowed.
    @see #pushBack()
  */
  private int now;
  
  /**
    Constructs a FastInput instance with default buffer size of 4kB
    and opens it with the given InputStream.
    @see #open(java.io.InputStream)
  */  
  public FastInput(InputStream is)
  {
    this(4096);
    open(is);
  }

  /**
    Constructs a FastInput instance with default buffer size of 4kB.
    You must call open before using this instance.
    @see #open(java.io.InputStream)
  */  
  public FastInput()
  {
    this(4096);
  }

  /**
    Constructs a FastInput instance with custom buffer size.
    You must call open before using this instance.
    @see #open(java.io.InputStream)
  */  
  public FastInput(int nbuf)
  {
    this.nbuf=nbuf;
    buf_bak=new byte[nbuf];
    blocksread=Integer.MIN_VALUE;
    len=Integer.MIN_VALUE;
    now=-1;
  }
  
  /**
    Opens the FastInput object. This means, the FastInput object 
    is bound to an input stream.
    @throws IllegalArgumentException if the FastInput object is 
            already open.
    @see #close()
  */
  public final void open(InputStream is)
  {
    if(this.is!=null) throw new IllegalArgumentException();
    if(is==null) throw new IllegalArgumentException();
    if(blocksread!=Integer.MIN_VALUE) throw new IllegalArgumentException();
    if(len!=Integer.MIN_VALUE) throw new IllegalArgumentException();
    if(now!=-1) throw new IllegalArgumentException();

    this.is=is;
    buf=buf_bak;
    blocksread=0;
    len=0;
    now=0;
  }
  
  /**
    Closes the input stream. Does nothing if already closed. A closed
    FastInput object may be opened with another input stream. In this
    case the allocated buffer is reused.
    @see #open(java.io.InputStream)
  */
  public final void close() throws IOException
  {
    if(is!=null)
    {
      is.close();
      is=null;
      buf=null;
      blocksread=Integer.MIN_VALUE;
      len=Integer.MIN_VALUE;
      now=-1;
    }
  }  
  
  /**
    The number bytes at least to be read on every readBuffer call.
    If there are less bytes available in the stream, readBuffer
    blocks.
    @see #readBuffer()
  */  
  static private final int minimumBytesToRead=10;
  
  /**
    Reads a new buffer.
    Do not call this method, when the last character in the
    current buffer has been read, but when the first character
    not in the current buffer gets read.
    Simply speaking, after calling readBuffer() you must read at least
    one character from buffer, so that you leave a non-zero buffer
    pointer.
    @see #pushBack()
  */  
  private final void readBuffer() throws IOException, EndException
  {
    //System.out.println("readBuffer() at "+getBytesRead());
    if(now!=len) throw new IllegalArgumentException();
    
    int rd=is.available();
    if(rd>nbuf) rd=nbuf;
    else if(rd<minimumBytesToRead) rd=minimumBytesToRead;
    
    rd=is.read(buf, 0, rd);
    if(rd<0) 
    { 
      close();
      throw new EndException(); 
    }
    if(rd==0) throw new IllegalArgumentException();
    
    blocksread+=len;
    len=rd;
    now=0;
  }
  
  public final int getBytesRead()
  {
    return now+blocksread;
  }
  
  /**
    Moves the file pointer back by one character. pushBack() is 
    available only for one character. Thus pushBack() is not available 
    immediatly after pushBack(), skipWhile() and construction.
    <P>
    There is no explicit pushback buffer. The buffer pointer (now) simply
    gets decreased by one. The class ensures, that the buffer
    pointer is not zero outside of public methods, whenever pushBack()
    is available.
    @see #skipWhile(byte)
  */
  public final void pushBack()
  {
    now--;
  }
    
  /**
    Reads a single character from the stream. 
    This method makes pushback available again.
    @see #pushBack()
  */
  public final char read() throws IOException, EndException
  {
    if(now>=len) readBuffer();
    return (char)(buf[now++]);
  }
  
  /**
    Reads a single character from the stream without moving the 
    file pointer. 
    <B>This method leaves no pushback buffer!</B>
    @see #pushBack()
  */
  public final char peek() throws IOException, EndException
  {
    if(now>=len) readBuffer();
    return (char)(buf[now]);
  }
  
  /**
    Skips input until the first occurence of the given character.
    The file pointer stops immediatly after this occurence.
    This method makes pushback available again.
    @see #pushBack()
  */  
  public final void skipUntil(byte c) throws IOException, EndException
  {
    while(true)
    {
      while(now<len)
        if(buf[now++]==c) return;
      readBuffer();
    }    
  }
  
  /**
    Skips input until the first occurence of one of the two given 
    characters. The file pointer stops immediatly after this occurence.
    @return the character matched
    This method makes pushback available again.
    @see #pushBack()
  */  
  public final char skipUntil(byte c1, byte c2) throws IOException, EndException
  {
    while(true)
    {
      byte x;
      while(now<len)
        if((x=buf[now++])==c1 || x==c2) return (char)x;
      readBuffer();
    }    
  }
  
  /**
    Skips input until the first occurence of one of the three given 
    characters. The file pointer stops immediatly after this occurence.
    @return the character matched
    This method makes pushback available again.
    @see #pushBack()
  */  
  public final char skipUntil(byte c1, byte c2, byte c3) throws IOException, EndException
  {
    while(true)
    {
      byte x;
      while(now<len)
        if((x=buf[now++])==c1 || x==c2 || x==c3) return (char)x;
      readBuffer();
    }    
  }
  
  /**
    Skips input until the first occurence of one of the four given 
    characters. The file pointer stops immediatly after this occurence.
    @return the character matched
    This method makes pushback available again.
    @see #pushBack()
  */  
  public final char skipUntil(byte c1, byte c2, byte c3, byte c4) throws IOException, EndException
  {
    while(true)
    {
      byte x;
      while(now<len)
        if((x=buf[now++])==c1 || x==c2 || x==c3 || x==c4) return (char)x;
      readBuffer();
    }    
  }
  
  /**
    Skips input for all subsequent occurences of the given character.
    The file pointer stops before at the first mismatch.
    <B>This method leaves no pushback buffer.</B> 
    @see #pushBack()
  */  
  public final void skipWhile(byte c) throws IOException, EndException
  {
    while(true)
    {
      while(now<len)
        if(buf[now]!=c) return; else now++;
      readBuffer();
    }    
  }
  
  /**
    The file pointer stops before the first character which not in the tree.
    <B>This method leaves no pushback buffer!</B>
    @see #pushBack()
  */
  public final CharNode traverse(CharNode n) throws IOException, EndException
  {
    CharNode z=null;
    
    do
    {
      z=n;
      n=n.next(read());
    }
    while(n!=null);
    
    pushBack();
    return z;
  }
  
  /**
    Return a string showing the file pointer as a pipeline character
    in the context of the characters in the stream.
    Useful for debugging.
  */
  public final String toString()
  {
    StringBuffer s=new StringBuffer();
    int a=now-10; if(a<0)a=0;
    int b=now+10; if(b>len)b=len;
    s.append('[');
    for(int i=a; i<now; i++) s.append((char)buf[i]);
    s.append('|');
    for(int i=now; i<b; i++) s.append((char)buf[i]);
    s.append(']');
    return new String(s);
  }
  
  /**
    This exception is thrown by any method accessing the 
    input stream, when the end of stream is reached. Whenever
    this exception is thrown, the FastInput object is also closed.
    @see #close()
  */
  public class EndException extends Exception
  {
    EndException()
    {
      super("end of stream reached");
    }
  } 
 
}
