/*
  Copyright (C) 1998 Ralf Wiebicke

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You'll find the GPL at http://www.gnu.org/copyleft/gpl.html
*/

package de.rw7.token;

import java.io.*;
import java.util.*;

/**
  Parses a HTML-stream (given as FastInput) into tokens. These tokens may be:
  <UL>
  <LI>Opening tags</LI>
  <LI>Closing tags</LI>
  <LI>Attributs of an opening tag</LI>
  <LI>Text beetween the tags converted to strings</LI>
  </UL>
  The parser tries to accept any dirty HTML. There are no exceptions
  for invalid HTML. 
  <UL>
  <LI>tag ::= &lt; [/] name param* &gt;</LI>
  <LI>name ::= (c)*</LI>
  <LI>param ::= n* key [ n* = n* value ]</LI>
  <LI>key ::= (c)*</LI>
  <LI>value := (c)* | " (c-")* " | ' (c-')* '</LI>
  <LI>n::= &lt;a white space&gt;</LI>
  <LI>c::= &lt;a non whitespace character&gt;</LI>
  </UL>
  The parser is specialized to work on only a few interesting tags
  over a large amount of HTML.
*/

public class HtmlTokenizer 
{
  public static final String PRESENT="present";

  public static final int MODE_ABORT = 1;
  public static final int MODE_PARSE_TEXT = 2;
  public static final int MODE_IGNORE_COMMENTS = 4;
  public static final int MODE_STRICT_TAGS = 8;
  public static final int MODE_STRICT_ATTR = 16;
  
  private int initialMode;
  
  /**
    The registry of the opening tags, the consumer is interested in.
  */
  private TagNode open;

  /**
    The registry of the closing tags, the consumer is interested in.
  */
  private TagNode clos;

  public HtmlTokenizer(int initialMode)
  {
    this.initialMode=initialMode;
    open=new TagNode();
    clos=new TagNode();
  }
  
  public void addOpeningTag(String name, int code, String[] attr)
  {
    open.addTag(name, code, attr);
  }
    
  public void addClosingTag(String name, int code)
  {
    clos.addTag(name, code, null);
  }
    
  public void addTag(String name, int code, String[] attr)
  {
    addOpeningTag(name, code, attr);
    addClosingTag(name, code);
  }
    
  public void removeOpeningTag(String name)
  {
    open.removeName(name);
  }
    
  public void removeClosingTag(String name)
  {
    clos.removeName(name);
  }
    
  public void removeTag(String name)
  {
    open.removeName(name);
    clos.removeName(name);
  }
    
  public void printTree(PrintStream o) 
  { 
    open.print(o, '-',0); 
    clos.print(o, '/',0); 
  }
  
  public void read(FastInput in, HtmlConsumer t) throws IOException, FastInput.EndException
  {
    boolean closingtag; // current tag ist closing/opening
    StringBuffer val=new StringBuffer(300); // attribute value
    TagNode tag; // node of tag name
    AttNode att; // node of attribute name
    char limval; // delimiter char of attribute value: space, apos or quote
    int mode=initialMode;

    String[][] attBufs=new String[10][];
    for(int i=0; i<attBufs.length; i++) attBufs[i]=new String[i];
    
    charL: do
    {
      // FP is somewhere but not inside a html tag
      if((mode&MODE_PARSE_TEXT)!=0)
      {
        t.onText(in);
        if(in.read()!='<') throw new IllegalArgumentException();
      }
      else 
        in.skipUntil((byte)'<');

      // FP is behind the "<" of a tag
      switch(in.read())
      {
      case '!': readComments(in, (mode&MODE_IGNORE_COMMENTS)!=0); continue charL;
      case '/': closingtag=true;  tag=clos; break;
      default:  closingtag=false; tag=open; in.pushBack(); break;
      };

      // FP is on the first character of the tag name
      tag=(TagNode)(in.traverse(tag));
      //System.out.println((closingtag?"tagC:":"tagO:")+tag.name);

      // FP is on the character causing the traversion to abort
      switch(in.read())
      {
      case '\n':
        if((mode&MODE_STRICT_TAGS)!=0) 
          throw new IllegalArgumentException("lf encountered in tag "+in.toString());
        continue charL;
      case ' ':
        in.skipWhile((byte)' ');
        break;
      case '>':
        in.pushBack();
        break;
      default: 
        if((mode&MODE_STRICT_TAGS)!=0) 
          throw new IllegalArgumentException("non registered tag encountered "+in.toString());
        in.skipUntil((byte)'>',(byte)'\n'); 
        continue charL;
      };
      
      if(tag.name==null)
      {
        if((mode&MODE_STRICT_TAGS)!=0) 
          throw new IllegalArgumentException("non registered tag encountered, see HtmlTokenizer.addTag() (2)");
        in.skipUntil((byte)'>',(byte)'\n');
        continue charL;
      }

      if(closingtag)
      {
        mode=t.onClosingTag(tag.code,tag.name);
        in.skipUntil((byte)'>',(byte)'\n');
        continue charL;
      }
      
      if(tag.att==null)
      {
        mode=t.onOpeningTag(tag.code,tag.name,null);
        in.skipUntil((byte)'>',(byte)'\n');
        continue charL;
      }
      
      String[] curratt=attBufs[tag.attLen];
      for(int i=0; i<curratt.length; i++) curratt[i]=null;
      
      paramL: do
      {
        // FP is on the first character of an attribute name 
        //    or on the ">" closing the tag

        //System.out.println("preatt:"+in.toString());
        att=(AttNode)(in.traverse(tag.att));
        //System.out.println("pstatt:"+att.name+in.toString());
        
        // FP is on the character causing the traversion to abort
        switch(in.read())
        {
        case '\n': 
          if((mode&MODE_STRICT_ATTR)!=0) 
            throw new IllegalArgumentException("lf encountered in attribute name "+in.toString());
          continue charL;
        case '>':
          if(att.name!=null) 
            curratt[att.number]=PRESENT;
          //System.out.println("attend:"+in.toString());
          break paramL;
        case '=':
          //System.out.println("attequ:"+in.toString());
          break;
        case ' ':
          //System.out.println("attspc:"+in.toString());
          in.skipWhile((byte)' ');
          switch(in.read())
          {
          case '\n': 
            if((mode&MODE_STRICT_ATTR)!=0) 
              throw new IllegalArgumentException("lf encountered in attribute name "+in.toString());
            continue charL;
          case '>':
            if(att.name!=null) 
              curratt[att.number]=PRESENT;
            //System.out.println("attenD:"+in.toString());
            break paramL;
          case '=':
            //System.out.println("atteqU:"+in.toString());
            break;
          default:
            in.pushBack();
            if(att.name!=null) 
              curratt[att.number]=PRESENT;
            //System.out.println("attnxT:"+in.toString());
            continue paramL;  
          }
          break;
        default:
          //System.out.println("attukn:"+in.toString());

          if((mode&MODE_STRICT_ATTR)!=0)
            throw new IllegalArgumentException("non registered attribute name encountered "+in.toString());
          att=null;  

          char bf=in.skipUntil((byte)'>',(byte)'\n',(byte)' ',(byte)'=');
          if(bf==' ')
          {
            in.skipWhile((byte)' ');
            bf=in.read();
          }  

          switch(bf)
          {
          case '\n': 
            if((mode&MODE_STRICT_ATTR)!=0) 
              throw new IllegalArgumentException("lf encountered in attribute name "+in.toString());
            continue charL;
          case '>': 
            break paramL;
          case '=':
            break;
          default: 
            in.pushBack();
            continue paramL;
          }
          break;
        }
        
        // FP is behind the "=" of an attribute
        in.skipWhile((byte)' ');
          
        limval=in.read();  
        if(limval!='\"'&&limval!='\'')
        {
          limval=' ';
          in.pushBack();
        }  

        char b;
        if(att!=null)
        {
          val.setLength(0);
          for(b=in.read(); 
              b!='\n'&&b!='>'&&b!=limval; 
              b=in.read())
            val.append(b);
          curratt[att.number]=new String(val);
        }
        else
          b=in.skipUntil((byte)'\n', (byte)'>', (byte)limval); 

        switch(b)
        {
        case '\n': continue charL;
        case '>': break paramL;
        }
         
        in.skipWhile((byte)' ');
      } 
      while(true); // paramL
      
      mode=t.onOpeningTag(tag.code, tag.name, curratt);
    } 
    while((mode&MODE_ABORT)==0); // charL
  }
  
  public void readComments(FastInput in, boolean ignoreComments) throws IOException, FastInput.EndException
  {
    if(in.read()=='-'&&in.read()=='-')
    {
      if(ignoreComments)
      {
        commentL: do
        {
          in.skipUntil((byte)'-');
          if(in.read()!='-') continue commentL;
          in.skipWhile((byte)' ');
          if(in.read()!='>') continue commentL;
          break commentL;
        } 
        while(true);
      }
    }
    else
      in.skipUntil((byte)'>');
  }
    
  public static final void test()
  {
    Node.test();
  }
    
} // class HtmlTokenizer


abstract class Node implements CharNode
{
  abstract Node produce();
  abstract void print(PrintStream o, char c, int level);
  
  final Node addName(String name)
  {
    int i=0;
    Node m=null, n=this;

    while(n!=null && i<name.length())
    {
      m=n;
      n=n.get(name.charAt(i++));
    }
    i--;
    
    while(i<name.length())
    {
      if(n==null) n=produce();
      m.set(name.charAt(i++),n);
      m=n; n=null;
    }
    
    if(m.name!=null) throw new IllegalArgumentException("can't add a tag twice");
    m.name=name;
    return m;
  }
    
  final void removeName(String name)
  {
    Node n=this;
  
    Vector v=new Vector();
    if(name.length()==0) throw new IllegalArgumentException("can't remove emty tag");

    for(int i=0; i<name.length(); i++)
    {
      n=n.get(name.charAt(i));
      if(n==null) throw new IllegalArgumentException("tag to be removed not in tree");
      v.addElement(n);
    }

    n.name=null;
    
    while(n.name==null&&n.empty()&&n!=this)
    {
      v.setSize(v.size()-1);
      Node m=n;
      n= (v.size()==0) ? this : (Node)v.lastElement();
      n.remove(m);
    }
  }
  
  protected Node[] children=new Node[37];
  String name;
  
  private final void set(char c, Node child)
  {
    int i=getIndex(c);
    if(i+1==children.length) throw new IllegalArgumentException("character not allowed:"+c);
    children[i]=child;
  }
  
  protected final Node get(char c)
  {
    return children[getIndex(c)];
  }
  
  public final CharNode next(char c)
  {
    return children[getIndex(c)];
  }
  
  private final boolean empty()
  {
    for(int i=0; i<children.length; i++)
      if(children[i]!=null)
        return false;
    return true;
  }
  
  private final void remove(Node x)
  {
    for(int i=0; i<children.length; i++)
      if(children[i]==x)
        { children[i]=null; return; }
    throw new IllegalArgumentException("treenode to be removed not in node");
  }
  
  /**
  '0'..'9' -> 0..9
  'A'..'Z' -> 10..35
  'a'..'z' -> 10..35
  sonst 36
  */
  private static final int[] indexTable = 
  {
    36,36,36,36,36,36,36,36,36,36,
    36,36,36,36,36,36,36,36,36,36,
    36,36,36,36,36,36,36,36,36,36,
    36,36,36,36,36,36,36,36,36,36,
    36,36,36,36,36,36,36,36,
    0,1,2,3,4,5,6,7,8,9,
    36,36,36,36,36,36,36,
    10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,
    36,36,36,36,36,36,
    10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35
  };
  
  private static final int getIndex(char c)
  {
    return (c<='z')?indexTable[c]:36;
  }  

  public static final void test()
  {
    char c='\0';
    for(; c< '0'; c++) if(indexTable[c]!=36)      throw new IllegalArgumentException(String.valueOf(c)+indexTable[c]);
    for(; c<='9'; c++) if(indexTable[c]!=c-'0')   throw new IllegalArgumentException(String.valueOf(c)+indexTable[c]);
    for(; c< 'A'; c++) if(indexTable[c]!=36)      throw new IllegalArgumentException(String.valueOf(c)+indexTable[c]);
    for(; c<='Z'; c++) if(indexTable[c]!=c-'A'+10)throw new IllegalArgumentException(String.valueOf(c)+indexTable[c]);
    for(; c< 'a'; c++) if(indexTable[c]!=36)      throw new IllegalArgumentException(String.valueOf(c)+indexTable[c]);
    for(; c<='z'; c++) if(indexTable[c]!=c-'a'+10)throw new IllegalArgumentException(String.valueOf(c)+indexTable[c]);
    if(c!=indexTable.length) throw new IllegalArgumentException();
  }

} // class Node
    
class TagNode extends Node
{
  /**
    The root node for all attributes of this tag.
  */
  AttNode att; 
  
  int attLen=-1;
  int code=-1;
  
  Node produce()
  {
    return new TagNode();
  }

  TagNode addTag(String name, int code, String[] attributes)
  {
    TagNode t=(TagNode)(addName(name));
    t.code=code;

    if(attributes!=null)
    {
      t.att=new AttNode(attributes);
      t.attLen=attributes.length;
    }
    return t;
  }

  final void print(PrintStream o, char c, int level)
  {
    for(int i=0; i<level; i++) o.print(' ');

    o.print(c);
    if(name!=null) o.print("----"+name+"#"+code);
    if(children==null) o.print("<");
    o.println();
    
    if(att!=null) att.print(o,'*',level+4);
    
    for(char i='0'; i<='9'; i++)
      if(get(i)!=null)
        get(i).print(o,i,level+1);
    for(char i='A'; i<='Z'; i++)
      if(get(i)!=null)
        get(i).print(o,i,level+1);
  }
  
}

class AttNode extends Node
{
  int number=-1;
  
  AttNode()
  {}

  AttNode(String[] attributes)
  {
    for(int i=0; i<attributes.length; i++)
      ((AttNode)(addName(attributes[i]))).number=i;
  }

  Node produce()
  {
    return new AttNode();
  }
  
  final void print(PrintStream o, char c, int level)
  {
    for(int i=0; i<level; i++) o.print(' ');

    o.print(c);
    if(name!=null) o.print("----"+name+"##"+number);
    if(children==null) o.print("<");
    o.println();
    
    for(char i='0'; i<='9'; i++)
      if(get(i)!=null)
        get(i).print(o,i,level+1);
    for(char i='A'; i<='Z'; i++)
      if(get(i)!=null)
        get(i).print(o,i,level+1);
  }
  
}
